@@ -64,6 +64,8 @@ var SkipLoggedOutTraceMsgEnvs = []string{
 // NOTE: keep this to one line, and 80 characters max
 var loggedOutTraceMsg = fmt.Sprintf("Setup tracing at %%s. To hide set %s=1", SkipLoggedOutTraceMsgEnvs[0])
 
+//go:generate go run github.com/matryer/moq -out frontend_mock.go . Frontend
+
 type Frontend interface {
 	// Run starts a frontend, and runs the target function.
 	Run(ctx context.Context, opts dagui.FrontendOpts, f func(context.Context) (cleanups.CleanupF, error)) error
@@ -114,6 +116,16 @@ type SidebarSection struct {
 	KeyMap []key.Binding
 }
 
+func (sec SidebarSection) Body(width int) string {
+	if sec.Content != "" {
+		return sec.Content
+	}
+	if sec.ContentFunc != nil {
+		return sec.ContentFunc(width)
+	}
+	return ""
+}
+
 // ShellHandler defines the interface for handling shell interactions
 type ShellHandler interface {
 	// Handle processes shell input
@@ -190,6 +202,12 @@ const (
 	kwColor     = termenv.ANSICyan
 	faintColor  = termenv.ANSIBrightBlack
 	moduleColor = termenv.ANSIMagenta
+
+	// filesync upload colors
+	bytesColor = termenv.ANSIGreen
+	kbColor    = bytesColor
+	mbColor    = termenv.ANSIYellow
+	bigColor   = termenv.ANSIRed
 )
 
 func (r *renderer) indent(out TermOutput, depth int) {
@@ -533,6 +551,7 @@ var metricsVerbosity = map[string]int{
 	telemetry.NetstatTxPackets:         3,
 	telemetry.LLMInputTokens:           1,
 	telemetry.LLMOutputTokens:          1,
+	telemetry.FilesyncWrittenBytes:     3,
 }
 
 func (r renderer) renderMetrics(out TermOutput, span *dagui.Span) {
@@ -563,6 +582,9 @@ func (r renderer) renderMetrics(out TermOutput, span *dagui.Span) {
 		r.renderMetric(out, metricsByName, telemetry.LLMOutputTokens, "Output Tokens", humanizeTokens)
 		r.renderMetric(out, metricsByName, telemetry.LLMInputTokensCacheReads, "Token Cache Reads", humanizeTokens)
 		r.renderMetric(out, metricsByName, telemetry.LLMInputTokensCacheWrites, "Token Cache Writes", humanizeTokens)
+
+		// Filesync Stats
+		r.renderMetric(out, metricsByName, telemetry.FilesyncWrittenBytes, "Written Bytes", colorizeBytes)
 	}
 }
 
@@ -640,6 +662,22 @@ func humanizeBytes(v int64) string {
 	return humanize.Bytes(uint64(v))
 }
 
+func colorizeBytes(v int64) string {
+	vh := humanizeBytes(v)
+	vs := strings.Split(vh, " ")
+	if len(vs) != 2 {
+		return vh
+	}
+	switch vs[1] {
+	case "B", "kB":
+		return termenv.String(vh).Foreground(kbColor).String()
+	case "MB":
+		return termenv.String(vh).Foreground(mbColor).String()
+	default:
+		return termenv.String(vh).Foreground(bigColor).String()
+	}
+}
+
 func humanizeTokens(v int64) string {
 	return humanize.Commaf(float64(v))
 }
